# -*- coding: utf-8 -*-

# Copyright © 2009/2023 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the "Software"), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""A JSON parser using funcparserlib.

The parser is based on [the JSON grammar][1].

  [1]: https://tools.ietf.org/html/rfc4627
"""

import re
import sys
from pprint import pformat
from re import VERBOSE
from typing import (
    List,
    Sequence,
    Optional,
    Tuple,
    Any,
    Dict,
    Match,
    TypeVar,
    Callable,
    Text,
    Union,
)

from funcparserlib.lexer import TokenSpec, make_tokenizer, Token, LexerError
from funcparserlib.parser import (
    maybe,
    many,
    finished,
    forward_decl,
    NoParseError,
    Parser,
    tok,
)

ENCODING = "UTF-8"
# noinspection SpellCheckingInspection
regexps = {
    "escaped": r"""
        \\                                  # Escape
          ((?P<standard>["\\/bfnrt])        # Standard escapes
        | (u(?P<unicode>[0-9A-Fa-f]{4})))   # uXXXX
        """,
    "unescaped": r"""
        [^"\\]                              # Unescaped: avoid ["\\]
        """,
}
re_esc = re.compile(regexps["escaped"], VERBOSE)
T = TypeVar("T")
JsonValue = Union[None, bool, dict, list, int, float, str]
JsonMember = Tuple[str, JsonValue]


def tokenize(s: str) -> List[Token]:
    specs = [
        TokenSpec("space", r"[ \t\r\n]+"),
        TokenSpec("string", r'"(%(unescaped)s | %(escaped)s)*"' % regexps, VERBOSE),
        TokenSpec(
            "number",
            r"""
            -?                  # Minus
            (0|([1-9][0-9]*))   # Int
            (\.[0-9]+)?         # Frac
            ([Ee][+-]?[0-9]+)?   # Exp
            """,
            VERBOSE,
        ),
        TokenSpec("op", r"[{}\[\]\-,:]"),
        TokenSpec("name", r"[A-Za-z_][A-Za-z_0-9]*"),
    ]
    useless = ["space"]
    t = make_tokenizer(specs)
    return [x for x in t(s) if x.type not in useless]


def parse(tokens: Sequence[Token]) -> JsonValue:
    def const(x: T) -> Callable[[Any], T]:
        return lambda _: x

    def op(s: str) -> Parser[Token, str]:
        return tok("op", s)

    def n(s: str) -> Parser[Token, Text]:
        return tok("name", s)

    def make_array(
        values: Optional[Tuple[JsonValue, List[JsonValue]]]
    ) -> List[JsonValue]:
        if values is None:
            return []
        else:
            return [values[0]] + values[1]

    def make_object(
        values: Optional[Tuple[JsonMember, List[JsonMember]]]
    ) -> Dict[str, Any]:
        if values is None:
            return {}
        else:
            first, rest = values
            k, v = first
            d = {k: v}
            d.update(rest)
            return d

    def make_number(s: str) -> Union[int, float]:
        try:
            return int(s)
        except ValueError:
            return float(s)

    def unescape(s: str) -> str:
        std = {
            '"': '"',
            "\\": "\\",
            "/": "/",
            "b": "\b",
            "f": "\f",
            "n": "\n",
            "r": "\r",
            "t": "\t",
        }

        def sub(m: Match[str]) -> str:
            if m.group("standard") is not None:
                return std[m.group("standard")]
            else:
                return chr(int(m.group("unicode"), 16))

        return re_esc.sub(sub, s)

    def make_string(s: str) -> str:
        return unescape(s[1:-1])

    def make_member(values: JsonMember) -> JsonMember:
        k, v = values
        return k, v

    null = n("null") >> const(None)
    true = n("true") >> const(True)
    false = n("false") >> const(False)
    number = tok("number") >> make_number
    string = tok("string") >> make_string
    value: Parser[Token, JsonValue] = forward_decl().named("json_value")
    member = string + -op(":") + value >> make_member
    json_object = (
        (-op("{") + maybe(member + many(-op(",") + member)) + -op("}")) >> make_object
    ).named("json_object")
    json_array = (
        (-op("[") + maybe(value + many(-op(",") + value)) + -op("]")) >> make_array
    ).named("json_array")
    value.define(null | true | false | json_object | json_array | number | string)
    json_text = value + -finished

    return json_text.parse(tokens)


def loads(s: str) -> JsonValue:
    return parse(tokenize(s))


def main() -> None:
    try:
        text = sys.stdin.read()
        tree = loads(text)
        print(pformat(tree))
    except (NoParseError, LexerError) as e:
        print("syntax error: %s" % e, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import unittest
from typing import Optional

from funcparserlib.parser import NoParseError
from funcparserlib.lexer import LexerError
from .dot import parse, tokenize, Graph, Edge, SubGraph, DefAttrs, Attr, Node


class DotTest(unittest.TestCase):
    def t(self, data: str, expected: Optional[Graph] = None) -> None:
        self.assertEqual(parse(tokenize(data)), expected)

    def test_comments(self) -> None:
        self.t(
            """
            /* комм 1 */
            graph /* комм 4 */ g1 {
                // комм 2 /* комм 3 */
            }
            // комм 5
        """,
            Graph(strict=None, type="graph", id="g1", stmts=[]),
        )

    def test_connected_subgraph(self) -> None:
        self.t(
            """
            digraph g1 {
                n1 -> n2 ->
                subgraph n3 {
                    nn1 -> nn2 -> nn3;
                    nn3 -> nn1;
                };
                subgraph n3 {} -> n1;
            }
        """,
            Graph(
                strict=None,
                type="digraph",
                id="g1",
                stmts=[
                    Edge(
                        nodes=[
                            "n1",
                            "n2",
                            SubGraph(
                                id="n3",
                                stmts=[
                                    Edge(nodes=["nn1", "nn2", "nn3"], attrs=[]),
                                    Edge(nodes=["nn3", "nn1"], attrs=[]),
                                ],
                            ),
                        ],
                        attrs=[],
                    ),
                    Edge(nodes=[SubGraph(id="n3", stmts=[]), "n1"], attrs=[]),
                ],
            ),
        )

    def test_default_attrs(self) -> None:
        self.t(
            """
            digraph g1 {
                page="3,3";
                graph [rotate=90];
                node [shape=box, color="#0000ff"];
                edge [style=dashed];
                n1 -> n2 -> n3;
                n3 -> n1;
            }
        """,
            Graph(
                strict=None,
                type="digraph",
                id="g1",
                stmts=[
                    DefAttrs(object="graph", attrs=[Attr(name="page", value='"3,3"')]),
                    DefAttrs(object="graph", attrs=[Attr(name="rotate", value="90")]),
                    DefAttrs(
                        object="node",
                        attrs=[
                            Attr(name="shape", value="box"),
                            Attr(name="color", value='"#0000ff"'),
                        ],
                    ),
                    DefAttrs(object="edge", attrs=[Attr(name="style", value="dashed")]),
                    Edge(nodes=["n1", "n2", "n3"], attrs=[]),
                    Edge(nodes=["n3", "n1"], attrs=[]),
                ],
            ),
        )

    def test_empty_graph(self) -> None:
        self.t(
            """
            graph g1 {}
        """,
            Graph(strict=None, type="graph", id="g1", stmts=[]),
        )

    def test_few_attrs(self) -> None:
        self.t(
            """
            digraph g1 {
                    n1 [attr1, attr2 = value2];
            }
        """,
            Graph(
                strict=None,
                type="digraph",
                id="g1",
                stmts=[
                    Node(
                        id="n1",
                        attrs=[
                            Attr(name="attr1", value=None),
                            Attr(name="attr2", value="value2"),
                        ],
                    )
                ],
            ),
        )

    def test_few_nodes(self) -> None:
        self.t(
            """
            graph g1 {
                n1;
                n2;
                n3
            }
        """,
            Graph(
                strict=None,
                type="graph",
                id="g1",
                stmts=[
                    Node(id="n1", attrs=[]),
                    Node(id="n2", attrs=[]),
                    Node(id="n3", attrs=[]),
                ],
            ),
        )

    def test_illegal_comma(self) -> None:
        try:
            self.t(
                """
                graph g1 {
                    n1;
                    n2;
                    n3,
                }
            """
            )
        except NoParseError:
            pass
        else:
            self.fail("must raise NoParseError")

    def test_null(self) -> None:
        try:
            self.t("")
        except NoParseError:
            pass
        else:
            self.fail("must raise NoParseError")

    def test_simple_cycle(self) -> None:
        self.t(
            """
            digraph g1 {
                n1 -> n2 [w=5];
                n2 -> n3 [w=10];
                n3 -> n1 [w=7];
            }
        """,
            Graph(
                strict=None,
                type="digraph",
                id="g1",
                stmts=[
                    Edge(nodes=["n1", "n2"], attrs=[Attr(name="w", value="5")]),
                    Edge(nodes=["n2", "n3"], attrs=[Attr(name="w", value="10")]),
                    Edge(nodes=["n3", "n1"], attrs=[Attr(name="w", value="7")]),
                ],
            ),
        )

    def test_single_unicode_char(self) -> None:
        try:
            self.t("ф")
        except LexerError:
            pass
        else:
            self.fail("must raise LexerError")

    def test_unicode_names(self) -> None:
        self.t(
            """
            digraph g1 {
                n1 -> "Медведь" [label="Поехали!"];
                "Медведь" -> n3 [label="Добро пожаловать!"];
                n3 -> n1 ["Водка"="Селёдка"];
            }
        """,
            Graph(
                strict=None,
                type="digraph",
                id="g1",
                stmts=[
                    Edge(
                        nodes=["n1", '"Медведь"'],
                        attrs=[Attr(name="label", value='"Поехали!"')],
                    ),
                    Edge(
                        nodes=['"Медведь"', "n3"],
                        attrs=[Attr(name="label", value='"Добро пожаловать!"')],
                    ),
                    Edge(
                        nodes=["n3", "n1"],
                        attrs=[Attr(name='"Водка"', value='"Селёдка"')],
                    ),
                ],
            ),
        )



# Copyright © 2009/2023 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the "Software"), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""A DOT language parser using funcparserlib.

The parser is based on [the DOT grammar][1]. It is pretty complete with a few
not supported things:

* String escapes
* Ports and compass points
* XML identifiers

At the moment, the parser builds only a parse tree, not an abstract syntax tree
(AST), or an API for dealing with DOT.

  [1]: https://www.graphviz.org/doc/info/lang.html
"""

import os
import sys
from re import MULTILINE
from typing import Sequence, List, TypeVar, Callable, NamedTuple, Union, Optional

from funcparserlib.lexer import TokenSpec, make_tokenizer, Token, LexerError
from funcparserlib.parser import (
    maybe,
    many,
    finished,
    oneplus,
    forward_decl,
    NoParseError,
    Parser,
    tok,
)
from funcparserlib.util import pretty_tree

ENCODING = "UTF-8"


class Graph(NamedTuple):
    strict: Optional[str]
    type: Optional[str]
    id: Optional[str]
    stmts: List["Statement"]


class SubGraph(NamedTuple):
    id: Optional[str]
    stmts: List["Statement"]


class Attr(NamedTuple):
    name: str
    value: Optional[str]


class Node(NamedTuple):
    id: str
    attrs: List[Attr]


class Edge(NamedTuple):
    nodes: List[Union[str, SubGraph]]
    attrs: List[Attr]


class DefAttrs(NamedTuple):
    object: str
    attrs: List[Attr]


Statement = Union[DefAttrs, Edge, SubGraph, Node]


T = TypeVar("T")


def tokenize(s: str) -> Sequence[Token]:
    specs = [
        TokenSpec("Comment", r"/\*(.|[\r\n])*?\*/", MULTILINE),
        TokenSpec("Comment", r"//.*"),
        TokenSpec("NL", r"[\r\n]+"),
        TokenSpec("Space", r"[ \t\r\n]+"),
        TokenSpec("Name", r"[A-Za-z\200-\377_][A-Za-z\200-\377_0-9]*"),
        TokenSpec("Op", r"[{};,=\[\]]|(->)|(--)"),
        TokenSpec("Number", r"-?(\.[0-9]+)|([0-9]+(\.[0-9]*)?)"),
        TokenSpec("String", r'"[^"]*"'),  # '\"' escapes are ignored
    ]
    useless = ["Comment", "NL", "Space"]
    t = make_tokenizer(specs)
    return [x for x in t(s) if x.type not in useless]


def parse(tokens: Sequence[Token]) -> Graph:
    def un_arg(f: Callable[..., T]) -> Callable[[tuple], T]:
        return lambda args: f(*args)

    def flatten(xs: List[List[Attr]]) -> List[Attr]:
        return sum(xs, [])

    def n(s: str) -> Parser[Token, str]:
        return tok("Name", s)

    def op(s: str) -> Parser[Token, str]:
        return tok("Op", s)

    dot_id = (tok("Name") | tok("Number") | tok("String")).named("id")

    def make_graph_attr(args: tuple) -> DefAttrs:
        return DefAttrs("graph", [Attr(*args)])

    def make_edge(
        node: Union[str, SubGraph], xs: List[Union[str, SubGraph]], attrs: List[Attr]
    ) -> Edge:
        return Edge([node] + xs, attrs)

    node_id = dot_id  # + maybe(port)
    a_list = dot_id + maybe(-op("=") + dot_id) + -maybe(op(",")) >> un_arg(Attr)
    attr_list = many(-op("[") + many(a_list) + -op("]")) >> flatten
    attr_stmt = (n("graph") | n("node") | n("edge")) + attr_list >> un_arg(DefAttrs)
    graph_attr = dot_id + -op("=") + dot_id >> make_graph_attr
    node_stmt = node_id + attr_list >> un_arg(Node)
    # We use a forward_decl because of circular definitions like
    # (stmt_list -> stmt -> subgraph -> stmt_list)
    subgraph: Parser[Token, SubGraph] = forward_decl()
    edge_rhs = -(op("->") | op("--")) + (subgraph | node_id)
    edge_stmt = (subgraph | node_id) + oneplus(edge_rhs) + attr_list >> un_arg(
        make_edge
    )
    stmt = attr_stmt | edge_stmt | subgraph | graph_attr | node_stmt
    stmt_list = many(stmt + -maybe(op(";")))
    graph_body = -op("{") + stmt_list + -op("}")
    subgraph.define(-n("subgraph") + maybe(dot_id) + graph_body >> un_arg(SubGraph))
    graph_modifiers = maybe(n("strict")) + maybe(n("graph") | n("digraph"))
    graph = graph_modifiers + maybe(dot_id) + graph_body >> un_arg(Graph)
    dotfile = graph + -finished

    return dotfile.parse(tokens)


def pretty_parse_tree(obj: object) -> str:
    class NamedValues(NamedTuple):
        name: str
        values: Sequence[object]

    def kids(x: object) -> Sequence[object]:
        if isinstance(x, (Graph, SubGraph)):
            return [NamedValues("stmts", x.stmts)]
        elif isinstance(x, (Node, DefAttrs)):
            return [NamedValues("attrs", x.attrs)]
        elif isinstance(x, Edge):
            return [NamedValues("nodes", x.nodes), NamedValues("attrs", x.attrs)]
        elif isinstance(x, NamedValues):
            return x.values
        else:
            return []

    def show(x: object) -> str:
        if isinstance(x, NamedValues):
            return x.name
        elif isinstance(x, Graph):
            return "Graph [id=%s, strict=%r, type=%s]" % (
                x.id,
                x.strict is not None,
                x.type,
            )
        elif isinstance(x, SubGraph):
            return "SubGraph [id=%s]" % (x.id,)
        elif isinstance(x, Edge):
            return "Edge"
        elif isinstance(x, Attr):
            return "Attr [name=%s, value=%s]" % (x.name, x.value)
        elif isinstance(x, DefAttrs):
            return "DefAttrs [object=%s]" % (x.object,)
        elif isinstance(x, Node):
            return "Node [id=%s]" % (x.id,)
        else:
            return str(x)

    return pretty_tree(obj, kids, show)


def main() -> None:
    # import logging
    # logging.basicConfig(level=logging.DEBUG)
    # import funcparserlib
    # funcparserlib.parser.debug = True
    try:
        stdin = os.fdopen(sys.stdin.fileno(), "rb")
        text = stdin.read().decode(ENCODING)
        tree = parse(tokenize(text))
        # print(pformat(tree))
        print(pretty_parse_tree(tree).encode(ENCODING))
    except (NoParseError, LexerError) as e:
        msg = ("syntax error: %s" % e).encode(ENCODING)
        print(msg, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import unittest
from typing import Optional

from funcparserlib.parser import NoParseError
from funcparserlib.lexer import LexerError
from . import json


class JsonTest(unittest.TestCase):
    def t(self, data: str, expected: Optional[object] = None) -> None:
        self.assertEqual(json.loads(data), expected)

    def test_1_array(self) -> None:
        self.t("[1]", [1])

    def test_1_object(self) -> None:
        self.t('{"foo": "bar"}', {"foo": "bar"})

    def test_bool_and_null(self) -> None:
        self.t("[null, true, false]", [None, True, False])

    def test_empty_array(self) -> None:
        self.t("[]", [])

    def test_empty_object(self) -> None:
        self.t("{}", {})

    def test_many_array(self) -> None:
        self.t("[1, 2, [3, 4, 5], 6]", [1, 2, [3, 4, 5], 6])

    def test_many_object(self) -> None:
        # noinspection SpellCheckingInspection
        self.t(
            """
            {
                "foo": 1,
                "bar":
                {
                    "baz": 2,
                    "quux": [true, false],
                    "{}": {}
                },
                "spam": "eggs"
            }
        """,
            {
                "foo": 1,
                "bar": {
                    "baz": 2,
                    "quux": [True, False],
                    "{}": {},
                },
                "spam": "eggs",
            },
        )

    def test_null(self) -> None:
        try:
            self.t("")
        except NoParseError:
            pass
        else:
            self.fail("must raise NoParseError")

    def test_numbers(self) -> None:
        self.t(
            """\
            [
                0, 1, -1, 14, -14, 65536,
                0.0, 3.14, -3.14, -123.456,
                6.67428e-11, -1.602176e-19, 6.67428E-11
            ]
        """,
            [
                0,
                1,
                -1,
                14,
                -14,
                65536,
                0.0,
                3.14,
                -3.14,
                -123.456,
                6.67428e-11,
                -1.602176e-19,
                6.67428e-11,
            ],
        )

    def test_strings(self) -> None:
        # noinspection SpellCheckingInspection
        self.t(
            r"""
            [
                ["", "hello", "hello world!"],
                ["привет, мир!", "λx.x"],
                ["\"", "\\", "\/", "\b", "\f", "\n", "\r", "\t"],
                ["\u0000", "\u03bb", "\uffff", "\uFFFF"],
                ["вот функция идентичности:\nλx.x\nили так:\n\u03bbx.x"]
            ]
        """,
            [
                ["", "hello", "hello world!"],
                ["привет, мир!", "λx.x"],
                ['"', "\\", "/", "\x08", "\x0c", "\n", "\r", "\t"],
                ["\u0000", "\u03bb", "\uffff", "\uffff"],
                ["вот функция идентичности:\nλx.x\nили так:\n\u03bbx.x"],
            ],
        )

    def test_toplevel_string(self) -> None:
        try:
            self.t("неправильно")
        except LexerError:
            pass
        else:
            self.fail("must raise LexerError")



# -*- coding: utf-8 -*-

import unittest
from typing import Optional, Tuple

from funcparserlib.lexer import TokenSpec, make_tokenizer, LexerError, Token
from funcparserlib.parser import (
    a,
    many,
    NoParseError,
    oneplus,
    Parser,
    maybe,
    _Ignored,  # noqa
    tok,
    finished,
    forward_decl,
    some,
)


class ParsingTest(unittest.TestCase):
    def test_oneplus(self) -> None:
        x = a("x")
        y = a("y")
        expr = oneplus(x + y)
        # noinspection SpellCheckingInspection
        self.assertEqual(expr.parse("xyxyxy"), ([("x", "y"), ("x", "y"), ("x", "y")]))

    # Issue 31
    def test_many_backtracking(self) -> None:
        x = a("x")
        y = a("y")
        expr = many(x + y) + x + x
        # noinspection SpellCheckingInspection
        self.assertEqual(expr.parse("xyxyxx"), ([("x", "y"), ("x", "y")], "x", "x"))

    # Issue 14
    def test_error_info(self) -> None:
        tokenize = make_tokenizer(
            [
                TokenSpec("keyword", r"\b(is|end)\b"),
                TokenSpec("id", r"[a-z_]+"),
                ("space", (r"[ \t]+",)),  # Legacy token spec
                TokenSpec("nl", r"[\n\r]+"),
            ]
        )
        with self.assertRaises(LexerError) as ctx:
            list(tokenize("f is ф"))
        self.assertEqual(str(ctx.exception), 'cannot tokenize data: 1,6: "f is \u0444"')

        def make_equality(values: Tuple[str, str]) -> Tuple[str, str]:
            v1, v2 = values
            return v1, v2

        tok_id = tok("id")
        equality = tok_id + -tok("keyword", "is") + tok_id >> make_equality
        expr = equality + -tok("nl")
        file = many(expr) + tok("keyword", "end")

        msg = """\
spam is eggs
foo is_not bar
end"""
        tokens = [x for x in tokenize(msg) if x.type != "space"]
        with self.assertRaises(NoParseError) as ctx2:
            file.parse(tokens)
        self.assertEqual(ctx2.exception.state.pos, 4)
        self.assertEqual(ctx2.exception.state.max, 5)
        # May raise KeyError
        t = tokens[ctx2.exception.state.max]
        self.assertEqual(t, Token("id", "is_not"))
        self.assertEqual((t.start, t.end), ((2, 5), (2, 10)))
        self.assertEqual(
            ctx2.exception.msg,
            "2,5-2,10: got unexpected token: 'is_not', expected: 'is'",
        )

    def test_ok_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, str] = -x + y
        self.assertEqual(expr.parse("xy"), "y")

    def test_ignored_ok(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, str] = x + -y
        self.assertEqual(expr.parse("xy"), "x")

    def test_ignored_ok_ok(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Tuple[str, str]] = -x + y + x
        self.assertEqual(expr.parse("xyx"), ("y", "x"))

    def test_ok_ignored_ok(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Tuple[str, str]] = x + -y + x
        self.assertEqual(expr.parse("xyx"), ("x", "x"))

    def test_ok_ok_ok(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Tuple[str, str]] = x + y + x
        self.assertEqual(expr.parse("xyx"), ("x", "y", "x"))

    def test_ok_ok_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Tuple[str, str]] = x + y + -x
        self.assertEqual(expr.parse("xyx"), ("x", "y"))

    def test_ignored_ignored_ok(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, str] = -x + -x + y
        self.assertEqual(expr.parse("xxy"), "y")

    def test_ok_ignored_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, str] = x + -y + -y
        self.assertEqual(expr.parse("xyy"), "x")

    def test_ignored_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, _Ignored] = -x + -y
        self.assertEqual(expr.parse("xy"), _Ignored("y"))

    def test_ignored_ignored_ignored(self) -> None:
        x = a("x")
        y = a("y")
        z = a("z")
        expr: Parser[str, _Ignored] = -x + -y + -z
        self.assertEqual(expr.parse("xyz"), _Ignored("z"))

    def test_ignored_maybe(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, str] = -maybe(x) + y
        self.assertEqual(expr.parse("xy"), "y")
        self.assertEqual(expr.parse("y"), "y")

    def test_maybe_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Tuple[Optional[_Ignored], str]] = maybe(-x) + y
        self.assertEqual(expr.parse("xy"), (_Ignored("x"), "y"))
        self.assertEqual(expr.parse("y"), (None, "y"))

    def test_ignored_maybe_ignored(self) -> None:
        x = a("x")
        y = a("y")
        expr: Parser[str, Optional[str]] = -x + maybe(y) + -x
        self.assertEqual(expr.parse("xyx"), "y")
        self.assertEqual(expr.parse("xx"), None)

    def test_compare_token_with_none(self) -> None:
        # https://github.com/vlasovskikh/funcparserlib/pull/58
        specs = [
            ("id", (r"\w+",)),
        ]
        tokenize = make_tokenizer(specs)
        tokens = list(tokenize("foo"))
        expr = maybe(a(None))
        self.assertEqual(expr.parse(tokens), None)  # type: ignore

    def test_seq_parse_error(self) -> None:
        expr = a("x") + a("y")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xz")
        self.assertEqual(ctx.exception.msg, "got unexpected token: 'z', expected: 'y'")

    def test_alt_2_parse_error(self) -> None:
        expr = a("x") + (a("x") | a("y"))
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xz")
        self.assertEqual(
            ctx.exception.msg, "got unexpected token: 'z', expected: 'x' or 'y'"
        )

    def test_alt_3_parse_error(self) -> None:
        expr = a("x") + (a("x") | a("y") | a("z"))
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xa")
        self.assertEqual(
            ctx.exception.msg,
            "got unexpected token: 'a', expected: 'x' or 'y' or 'z'",
        )

    def test_alt_3_two_steps_parse_error(self) -> None:
        expr = a("x") + (a("x") | (a("y") + a("a")))
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xyz")
        self.assertEqual(ctx.exception.msg, "got unexpected token: 'z', expected: 'a'")

    def test_expected_eof_error(self) -> None:
        expr = a("x") + finished
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xy")
        self.assertEqual(
            ctx.exception.msg,
            "got unexpected token: 'y', expected: end of input",
        )

    def test_expected_second_in_sequence_error(self) -> None:
        expr = a("x") + a("y")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xz")
        self.assertEqual(ctx.exception.msg, "got unexpected token: 'z', expected: 'y'")

    def test_forward_decl_nested_matching_error(self) -> None:
        expr = forward_decl()
        expr.define(a("x") + maybe(expr) + a("y"))
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("xxy")
        self.assertEqual(
            ctx.exception.msg, "got unexpected end of input, expected: 'y'"
        )

    def test_expected_token_type_error(self) -> None:
        expr = tok("number")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse([Token("id", "x")])
        self.assertEqual(
            ctx.exception.msg, "got unexpected token: 'x', expected: number"
        )

    def test_expected_exact_token_error(self) -> None:
        expr = tok("operator", "=")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse([Token("operator", "+")])
        self.assertEqual(ctx.exception.msg, "got unexpected token: '+', expected: '='")

    def test_unexpected_eof(self) -> None:
        expr = (a("x") + a("y")) | a("z")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("x")
        self.assertEqual(
            ctx.exception.msg, "got unexpected end of input, expected: 'y'"
        )

    def test_expected_transform_parsing_results_error(self) -> None:
        expr = (a("1") >> int) | a("2")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("x")
        self.assertEqual(
            ctx.exception.msg, "got unexpected token: 'x', expected: '1' or '2'"
        )

    def test_expected_sequence_with_skipped_parts(self) -> None:
        expr = (-a("x") + a("y")) | a("z")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("b")
        self.assertEqual(
            ctx.exception.msg,
            "got unexpected token: 'b', expected: ('x', 'y') or 'z'",
        )

    def test_expected_some_without_name(self) -> None:
        def lowercase(t: str) -> bool:
            return t.islower()

        expr = some(lowercase)
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("A")
        self.assertEqual(
            ctx.exception.msg, "got unexpected token: 'A', expected: some(...)"
        )

    def test_expected_forward_decl_without_name(self) -> None:
        nested = forward_decl()
        nested.define(-a("a") + maybe(nested) + -a("z"))
        expr = nested | a("x")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("y")
        self.assertEqual(
            ctx.exception.msg,
            "got unexpected token: 'y', "
            "expected: (('a', [ forward_decl() ]), 'z') or 'x'",
        )

    def test_expected_forward_decl_with_name(self) -> None:
        nested = forward_decl().named("nested")
        nested.define(-a("a") + maybe(nested) + -a("z"))
        expr = nested | a("x")
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("y")
        self.assertEqual(
            ctx.exception.msg,
            "got unexpected token: 'y', expected: (('a', [ nested ]), 'z') or 'x'",
        )

    def test_end_of_input_after_many_alternatives(self) -> None:
        brackets = a("[") + a("]")
        expr = many(a("x") | brackets) + finished
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("[")
        self.assertEqual(
            ctx.exception.msg, "got unexpected end of input, expected: ']'"
        )

    def test_parse_one_more_then_rollback_to_single(self) -> None:
        mul = a("x") + many(a("*") + a("y"))
        add = mul + many(a("+") + mul)
        expr = add + finished
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("x*")
        self.assertEqual(
            ctx.exception.msg, "got unexpected end of input, expected: 'y'"
        )

    def test_parse_one_more_then_rollback_to_alternative(self) -> None:
        mul = a("x") + many(a("*") + a("y"))
        addsub = mul + many((a("+") | a("-")) + mul)
        expr = addsub + finished
        with self.assertRaises(NoParseError) as ctx:
            expr.parse("x*")
        self.assertEqual(
            ctx.exception.msg, "got unexpected end of input, expected: 'y'"
        )
